"""
Single-threaded job queue. This should be turned into a parallel job queue at some point.
"""
class JobQueue:
    def __init__(self, name=None, parent=None) -> None:
        self.name = name if name else ""
        self.parent = parent
        self.waited = False
        self.subs = []
        self.logs = []

    def __del__(self):
        if not self.waited and len(self.subs) > 0:
            print(f'X[{self._name()}]: Failed to wait for subordinate jobs!', flush=True)
        if self.logs:
            print(f'X[{self._name()}]: Failed to print logs!', flush=True)

    def submit(self, name, job):
        sub = JobQueue(name=name, parent=self)
        self.subs.append(sub)
        job(sub)
        for level, msg in sub.logs:
            print(f'{level}[{sub._name()}]: {msg}', flush=True)
        sub.logs = []
        pass

    def error(self, msg):
        self._log("E", msg)

    def info(self, msg):
        self._log("I", msg)

    def warning(self, msg):
        self._log("W", msg)

    def _log(self, level, msg):
        self.logs.append((level, msg))

    def wait(self):
        self.waited = True
        pass

    def _name(self):
        if self.parent:
            return f'{self.parent._name()}/{self.name}'
        return self.name



import os
import time
import subprocess
from subprocess import Popen, TimeoutExpired
from threading import Thread
from typing import List

BUFFER_SIZE = 128 * 1024

class ClosedException(BaseException):
    pass

"""
We don't need the full complexity of the python streams, so let's work directly off OS file descriptors.
"""
class BasicStream:
    def __init__(self, fd) -> None:
        self.fd = fd
    def read(self, n):
        if self.fd == -1:
            raise ClosedException("Read on closed stream")
        return os.read(self.fd, n)
    def write(self, data):
        if self.fd == -1:
            raise ClosedException("Write on closed stream")
        return os.write(self.fd, data)
    def close(self):
        if self.fd != -1:
            try:
                os.close(self.fd)
            except OSError:
                pass
        self.fd = -1

"""
Create a subprocess and ensure that it's always making forward progress by consuming stdin.
"""
def subprocess_forward_progress(input, args, executable, timeout=10, progress_callback=None) -> List[str]:
    process = Popen(args=args, executable=executable, stdin=subprocess.PIPE, stdout=subprocess.DEVNULL, stderr=subprocess.PIPE)
    stdin = BasicStream(process.stdin.fileno())
    stderr = BasicStream(process.stderr.fileno())
    errors = []

    stderr_buffer = [b'']
    def stderr_reader(stderr, stderr_buffer):
        try:
            while True:
                r = stderr.read(128)
                stderr_buffer[0] += r
        except ClosedException:
            pass
        except:
            stderr.close()
    t1 = Thread(target=stderr_reader, args=(stderr, stderr_buffer,))
    t1.start()

    last_write = [time.monotonic()]
    progress = [0]
    def stdin_writer(stdin, progress, last_write):
        try:
            with open(input, 'rb') as f:
                fl = f.seek(0, 2)
                f.tell()
                f.seek(0, 0)
                w = 0
                while True:
                    bytes = f.read(BUFFER_SIZE)
                    if not bytes:
                        break
                    w += len(bytes)
                    stdin.write(bytes)
                    last_write[0] = time.monotonic()
                    progress[0] = w / fl
        except BrokenPipeError:
            errors.append("Process failed to read the entire input")
        except ClosedException:
            pass
        finally:
            stdin.close()
    t2 = Thread(target=stdin_writer, args=(stdin, progress, last_write,))
    t2.start()

    try:
        while True:
            if progress_callback:
                try:
                    progress_callback(progress[0])
                except:
                    pass
            if time.monotonic() - last_write[0] > timeout:
                errors.append("Process timed out reading from input stream")
                break
            if t1:
                t1.join(timeout=0)
                if not t1.is_alive():
                    t1 = None
            if t2:
                t2.join(timeout=0)
                if not t2.is_alive():
                    t2 = None
            try:
                ret = process.wait(timeout=0.01)
                if ret != 0:
                    errors.append(f"Process failed with exit code {ret}")
                break
            except TimeoutExpired:
                pass
            except KeyboardInterrupt as e:
                errors.append("Interrupted by user")
                raise e
            except:
                errors.append("Failed for unknown reason")
                break
    finally:
        stdin.close()
        stderr.close()
        Thread(target=lambda: process.kill()).start()
        if t1:
            t1.join()
        if t2:
            t2.join()
        if progress_callback:
            try:
                progress_callback(progress[1])
            except:
                pass
        if len(stderr_buffer[0]) > 0:
            errors.append("Process wrote to error stream: " + str(stderr_buffer[0], encoding='utf8', errors='replace'))

    return errors



from concurrent.futures import process
from pathlib import Path
import argparse
import re
import shlex
import sys

from .path_scan import PathScanner
from .jobqueue import JobQueue
from .ffmpeg_validator import FFMPEGValidateOperation, FFMPEG_SUPPORTED_EXTENSIONS
from .par2 import CreatePar2Operation, VerifyPar2Operation, DEFAULT_PAR2_CREATE_ARGS, DEFAULT_PAR2_VERIFY_ARGS
from .operation import Operation, PrintFilesOperation

"""
Default precious extensions that we want to preserve w/PAR2.
"""
DEFAULT_EXTENSIONS = ','.join(FFMPEG_SUPPORTED_EXTENSIONS + [
    # Docs
    "pdf",
    # Chiptunes
    "d64", "mod", "s3m"])
DEFAULT_IGNORE_FILES = ','.join([r"\.DS_Store", r"Thumbs\.db", r"\._.*", r".*\.par2", r".*\.filelist"])
DEFAULT_SPAM_FILES =','.join(["RARBG.txt", "RARBG_DO_NOT_MIRROR.exe", "WWW.YIFY-TORRENTS.COM.jpg", "www.YTS.AM.jpg", "WWW.YTS.TO.jpg", "www.YTS.LT.jpg"])

def process_dir(q: JobQueue, scanner: PathScanner, dir: Path, op: Operation):
    results = scanner.scan(dir)
    if results.unknown_extensions:
        q.warning(f"Unknown extensions found in path: {' '.join(results.unknown_extensions)}")
    if results.media_list:
        op.operate(q, dir, results.media_list)
    for dir in results.directory_list:
        q.submit(dir.name, lambda q: process_dir(q, scanner, dir, op))
    q.wait()

def compile_extension_regex(extensions):
    return re.compile('|'.join([f'\.{e}' for e in extensions.split(',')]), flags=re.IGNORECASE)

def compile_ignore_regex(files):
    return re.compile('|'.join([f'({f})' for f in files.split(',')]))

def do_main():
    parser = argparse.ArgumentParser(description='Process media directories to validate and add parity files')
    parser.add_argument("--hidden-container-prefix", dest="container_prefix", action="store", help=argparse.SUPPRESS)
    parser.add_argument("--hidden-container-pwd", dest="container_pwd", action="store", help=argparse.SUPPRESS)
    parser.add_argument("--root", required=True, dest="root_dir", action="store", help="root directory for media")
    parser.add_argument("--extensions", default=DEFAULT_EXTENSIONS, help=f"file extensions to include in processing (default {DEFAULT_EXTENSIONS})")
    parser.add_argument("--ignore", default=DEFAULT_IGNORE_FILES, help=f"file regular expressions to completely exclude in processing (default {DEFAULT_IGNORE_FILES})")
    commands = parser.add_subparsers(dest="command", required=True, help="sub-command help (use sub-command --help for more info)")
    verify_cmd = commands.add_parser("verify", help="verify media files are corruption-free with FFMPEG")
    print_cmd = commands.add_parser("print", help="print all media files")
    par2_create_cmd = commands.add_parser("par2-create", help="create a PAR2 archive in each directory")
    par2_create_cmd.add_argument("--par2-args", default=DEFAULT_PAR2_CREATE_ARGS, help=f"arguments to pass to PAR2 (default {DEFAULT_PAR2_CREATE_ARGS})")
    par2_create_cmd.add_argument("--name", dest="par2_name", default="recovery", help="recovery filename (for .par2 and .filelist files)")
    par2_verify_cmd = commands.add_parser("par2-verify", help="verify the PAR2 archive in each directory")
    par2_verify_cmd.add_argument("--par2-args", default=DEFAULT_PAR2_VERIFY_ARGS, help=f"arguments to pass to PAR2 (default {DEFAULT_PAR2_VERIFY_ARGS})")
    par2_verify_cmd.add_argument("--name", dest="par2_name", default="recovery", help="recovery filename (for .par2 and .filelist files)")

    args = parser.parse_args()

    extension_regex = compile_extension_regex(args.extensions)
    ignore_regex = compile_ignore_regex(args.ignore)
    if args.command == 'verify':
        operation = FFMPEGValidateOperation()
    if args.command == 'print':
        operation = PrintFilesOperation()
    if args.command == 'par2-create':
        operation = CreatePar2Operation(shlex.split(args.par2_args), args.par2_name)
    if args.command == 'par2-verify':
        operation = VerifyPar2Operation(shlex.split(args.par2_args), args.par2_name)
    q = JobQueue()

    # Docker-awareness
    root = Path(args.root_dir)
    if args.container_prefix:
        pwd = Path(args.container_pwd)
        root = pwd / root
        for candidate in args.container_prefix.split(','):
            candidate = Path(candidate) / root.relative_to('/')
            if candidate.exists():
                root = candidate
                break

    if not root.is_dir():
        print(f"Root must be a directory: {root}")
        sys.exit(1)

    scanner = PathScanner(
        supported_extension_matcher=lambda p: p.suffix and extension_regex.fullmatch(p.suffix),
        ignored_pattern_matcher=lambda p: ignore_regex.fullmatch(p.name),
        spam_files_matcher=lambda _: False)

    q.submit(None, lambda q: process_dir(q, scanner, root, operation))
    q.wait()

def main():
    try:
        do_main()
    except KeyboardInterrupt:
        print("Interrupted by user!")



import os

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from stat import S_ISDIR, S_ISREG
from typing import Callable, List

class EntryType(Enum):
    DIR = 0,
    FILE_MEDIA = 1,
    FILE_IGNORE = 2,
    FILE_SPAM = 3

@dataclass
class PathScanResults:
    directory_list: List[Path]
    media_list: List[Path]
    unknown_extensions: List[str]

class PathScanner:
    def __init__(self,
        supported_extension_matcher: Callable[[Path], bool] = None,
        ignored_pattern_matcher: Callable[[Path], bool] = None,
        spam_files_matcher: Callable[[Path], bool] = None) -> None:

        self.supported_extension_matcher = supported_extension_matcher
        self.ignored_pattern_matcher = ignored_pattern_matcher
        self.spam_files_matcher = spam_files_matcher

    def scan(self, dir):
        files = []
        dirs = []
        unknown_extensions = set()
        for f in os.listdir(dir):
            filename = dir / f
            if self.ignored_pattern_matcher(filename):
                continue
            st = os.lstat(filename)
            # Ignore zero-length files
            if st.st_size == 0:
                continue
            if S_ISREG(st.st_mode):
                if self.supported_extension_matcher(filename):
                    files.append(filename)
                else:
                    if filename.suffix:
                        unknown_extensions.add(filename.suffix)
            elif S_ISDIR(st.st_mode):
                dirs.append(filename)
            pass
        unknown_extensions = list(unknown_extensions)
        unknown_extensions.sort()
        files.sort()
        dirs.sort()

        return PathScanResults(directory_list=dirs, media_list=files, unknown_extensions=unknown_extensions)



from abc import abstractmethod

class Operation:
    @abstractmethod
    def operate(dir, files):
        pass

class PrintFilesOperation:
    def operate(self, q, dir, files):
        q.info(f"{len(files)} file(s)")
        for file in files:
            q.info(f"  {file.name}")
        pass



import os
import subprocess

from enum import Enum
from pathlib import Path
from typing import List
from subprocess import Popen

from .operation import Operation

DEFAULT_PAR2_CREATE_ARGS = ' '.join(['-u', '-n3', '-r10'])
DEFAULT_PAR2_VERIFY_ARGS = ' '.join(['-N'])

RECOVERY_LIST_HEADER = "[media-tools-v1]"

class RecoveryList:
    def __init__(self, files) -> None:
        self.files = files

    def read(file: Path):
        with open(file, 'rt') as f:
            header = f.readline()
            if header.rstrip() != RECOVERY_LIST_HEADER:
                raise Exception(f"Invalid {file.name} found (header was {header}), cannot create parity files")
            old_files = [x.rstrip() for x in f.readlines()]
            old_files.sort()
        return RecoveryList(old_files)

    def write(self, file: Path):
        with open(file, 'wt') as f:
            f.write("[media-tools-v1]\n")
            f.write('\n'.join([x.name for x in self.files]))

class RecoveryListState(Enum):
    MISSING = 0
    UP_TO_DATE = 1
    ERROR = 2

class Par2Operation(Operation):
    def __init__(self, args, recovery_name) -> None:
        self.args = list(args)
        self.recovery_name = recovery_name

    def recovery_list(self, dir):
        return dir / f'{self.recovery_name}.filelist'

    def par2_index(self, dir):
        return dir / f'{self.recovery_name}.par2'

    def validate_recovery_list(self, q, dir: Path, files: List[Path]) -> RecoveryListState:
        if not self.par2_index(dir).exists():
            if self.recovery_list(dir).exists():
                q.warning("File list exists, but PAR2 does not exist")
            return RecoveryListState.MISSING
        if self.recovery_list(dir).exists():
            try:
                list = RecoveryList.read(self.recovery_list(dir))
                if list.files != [x.name for x in files]:
                    q.warning("PAR2 exists, but is out-of-date")
                    q.warning(list.files)
                    q.warning([x.name for x in files])
                else:
                    q.info("PAR2 exists, and is up-to-date")
                    return RecoveryListState.UP_TO_DATE
            except Exception as e:
                q.error(str(e))
                return RecoveryListState.ERROR
        return RecoveryListState.MISSING

    def run_par2(self, q, dir, args):
        # q.info(args)
        cmd = Popen(args, executable="par2", encoding="utf8", errors="", cwd=dir, stdout=subprocess.PIPE, stdin=subprocess.DEVNULL, stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate()
        if stderr:
            q.error(stderr)
        elif not stdout.find("\nDone"):
            q.error(stdout)
        elif cmd.returncode != 0:
            q.error(f"PAR2 returned a non-zero errorcode: {cmd.returncode}")
        else:
            return True
        return False

class CreatePar2Operation(Par2Operation):
    def operate(self, q, dir, files):
        if self.validate_recovery_list(q, dir, files) != RecoveryListState.MISSING:
            return
        args = ['par2', 'c'] + self.args + ['--', f'{self.recovery_name}'] + [x.name for x in files]
        if self.run_par2(q, dir, args):
            if self.par2_index(dir).exists():
                RecoveryList(files).write(self.recovery_list(dir))
                q.info("Done")
            else:
                q.warning("No PAR2 files were generated")

class VerifyPar2Operation(Par2Operation):
    def operate(self, q, dir, files):
        if self.validate_recovery_list(q, dir, files) != RecoveryListState.UP_TO_DATE:
            q.warning("Unable to verify directory")
            return
        args = ['par2', 'v'] + self.args + ['--', f'{self.recovery_name}'] + [x.name for x in files]
        if self.run_par2(q, dir, args):
            # If the user requested removal of PAR2 files, we should unlink our recovery list too
            if not self.par2_index(dir).exists():
                q.warning("PAR2 files were removed after verification")
                os.unlink(self.recovery_list(dir))



import os
import re

from .jobqueue import JobQueue
from .forward_progress import subprocess_forward_progress
from .operation import Operation

def compile_extension_regex(*extensions):
    return re.compile('|'.join([f'\.{e}' for e in extensions]), flags=re.IGNORECASE)

FFMPEG_SUPPORTED_EXTENSIONS = [
    # Video
    "mp4", "mov", "avi", "mkv", "srt", "mpeg", "mpg",
    # Music
    "aac", "mp3", "flac", "ogg", "m4a", "wav", "wma",
    # Pics
    "jpg", "jpeg", "png", "gif"]
FFMPEG_SUPPORTED_EXTENSIONS_REGEX = compile_extension_regex(*FFMPEG_SUPPORTED_EXTENSIONS)

class ClosedException(BaseException):
    pass

"""
We don't need the full complexity of the python streams, so let's work directly off OS file descriptors.
"""
class BasicStream:
    def __init__(self, fd) -> None:
        self.fd = fd
    def read(self, n):
        if self.fd == -1:
            raise ClosedException("Read on closed stream")
        return os.read(self.fd, n)
    def write(self, data):
        if self.fd == -1:
            raise ClosedException("Write on closed stream")
        return os.write(self.fd, data)
    def close(self):
        if self.fd != -1:
            try:
                os.close(self.fd)
            except OSError:
                pass
        self.fd = -1

def ffmpeg_supports(filename):
    if FFMPEG_SUPPORTED_EXTENSIONS_REGEX.match(filename.suffix):
        return True
    return False

def ffmpeg_validate(input, timeout=10, executable="ffmpeg", progress_callback=None):
    args = [
        "-xerror",
        "-v", "error",
        "-i", "-",
        "-f", "null",
        "-"
    ]

    return subprocess_forward_progress(input, args, executable, timeout=timeout, progress_callback=progress_callback)

class FFMPEGValidateOperation(Operation):
    def operate(self, q: JobQueue, dir, files):
        stats = { 'good': 0, 'bad': 0, 'ignored': 0 }
        for file in files:
            q.submit(file.name, lambda q: self._job(q, stats, file))
        q.wait()
        if stats['ignored']:
            q.info(f"{stats['good']} good file(s), {stats['bad']} bad file(s), {stats['ignored']} ignored file(s)")
        elif stats['bad']:
            q.info(f"{stats['good']} good file(s), {stats['bad']} bad file(s)")
        else:
            q.info(f"{stats['good']} good file(s)")

    def _job(self, q, stats, file):
        if ffmpeg_supports(file):
            errors = ffmpeg_validate(file)
            if errors:
                stats['bad'] += 1
                q.error(errors)
            else:
                stats['good'] += 1
        else:
            stats['ignored'] += 1

if __name__ == '__main__':
    import sys
    errors = ffmpeg_validate(sys.argv[1], timeout=1, executable="ffmpeg")
    if errors:
        print(errors)
    else:
        print("Ok!")
